package com.laguerre.kan;

public final class Light extends Behaviour {
	public static enum RenderingMode {
		IMPORTANT,
		NOT_IMPORTANT
	}

	public static enum LightType {
		DIRECTIONAL,
		POINT,
		SPOT
	}

	private LightType lightType = LightType.POINT;
	private RenderingMode renderingMode = RenderingMode.NOT_IMPORTANT;
	private final Color diffuseColor = Color.getNewBlack();
	private final Color specularColor = Color.getNewBlack();
	private float innerCutOffAngleDeg = 10f;
	private float outerCutOffAngleDeg = 15f;
	/**
	 * The reference of this array will be passed and used directly by shaders.
	 * For the sake of performance, the following values are packed here:
	 * <ul>
	 *   <li>cosine inner angle (index 0), computed from <code>innerCutOffAngleDeg</code></li>
	 *   <li>cosine outer angle (index 1), computed from <code>outerCutOffAngleDeg</code></li>
	 * </ul>
	 */
	private final float[] cosineAngleValues = { 0f, 0f };

	/**
	 * The reference of this array will be passed and used directly by shaders.
	 * For the sake of performance, the following values are packed here:
	 * <ul>
	 *   <li>range (index 0)</li>
	 *   <li>constant attenuation (index 1)</li>
	 *   <li>linear attenuation (index 2)</li>
	 *   <li>quadratic attenuation (index 3)</li>
	 * </ul>
	 */
	private final float[] attenuationValues = { -1f, 1f, 0f, 0f };

	Light() {
	}

	public float[] getSharedCosineAngleValues() {
		return this.cosineAngleValues;
	}

	public float[] getSharedAttenuationValues() {
		return this.attenuationValues;
	}

	/**
	 * Memory placeholder of the light direction computed by <code>getSharedDirection</code>.
	 * This avoids many memory allocations.
	 */
	private final Vector3 lightDirection = Vector3.getNewOneVec();

	public Vector3 getSharedDirection() {
		// The user is asked to imagine that the ray light goes forward but
		// we use "backward" ray to get an inverse a ray light for shaders.
		getSharedTransform().getSharedWorldRotation().rotateDirection(
				this.lightDirection, Vector3.BACKWARD);
		return this.lightDirection;
	}

	public void setLightType(LightType lightType) {
		this.lightType = lightType;
	}

	public LightType getLightType() {
		return this.lightType;
	}

	public void setRenderingMode(RenderingMode renderingMode) {
		this.renderingMode = renderingMode;
	}

	public RenderingMode getRenderingMode() {
		return this.renderingMode;
	}

	public void setDiffuseColor(float r, float g, float b, float a) {
		this.diffuseColor.setValues(r, g, b, a);
	}

	public void setDiffuseColor(Color c) {
		this.diffuseColor.setValues(c);
	}

	public Color getSharedDiffuseColor() {
		return this.diffuseColor;
	}

	public void setSpecularColor(float r, float g, float b, float a) {
		this.specularColor.setValues(r, g, b, a);
	}

	public void setSpecularColor(Color c) {
		this.specularColor.setValues(c);
	}

	public Color getSharedSpecularColor() {
		return this.specularColor;
	}

	public void setRange(float range) {
		this.attenuationValues[0] = range;
	}

	public float getRange() {
		return this.attenuationValues[0];
	}

	public void setConstantAttenuation(float constantAttenuation) {
		this.attenuationValues[1] = constantAttenuation;
	}

	public float getConstantAttenuation() {
		return this.attenuationValues[1];
	}

	public void setLinearAttenuation(float linearAttenuation) {
		this.attenuationValues[2] = linearAttenuation;
	}

	public float getLinearAttenuation() {
		return this.attenuationValues[2];
	}

	public void setQuadraticAttenuation(float quadraticAttenuation) {
		this.attenuationValues[3] = quadraticAttenuation;
	}

	public float getQuadraticAttenuation() {
		return this.attenuationValues[3];
	}

	public void setInnerCutOffAngle(float innerCutOffAngleDeg) {
		this.innerCutOffAngleDeg = innerCutOffAngleDeg;
		this.cosineAngleValues[0] = (float) Math.cos(Math.toRadians(innerCutOffAngleDeg));
	}

	public float getInnerCutOffAngle() {
		return this.innerCutOffAngleDeg;
	}

	public void setOuterCutOffAngle(float outerCutOffAngleDeg) {
		this.outerCutOffAngleDeg = outerCutOffAngleDeg;
		this.cosineAngleValues[1] = (float) Math.cos(Math.toRadians(outerCutOffAngleDeg));
	}

	public float getOuterCutOffAngle() {
		return this.outerCutOffAngleDeg;
	}
}
