package com.laguerre.kan;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.ShortBuffer;
import java.util.ArrayList;

public class Mesh extends Component {
	public static final int POSITION_DATA_SIZE = 4;
	public static final int NORMAL_DATA_SIZE = 3;
	public static final int TANGENT_DATA_SIZE = 4;
	public static final int UV_DATA_SIZE = 2;

	protected static final class SubMeshData {
		public String name = "";
		public float[] vertices = new float[0];
		public float[] normals = new float[0];
		public float[] tangents = new float[0];
		public float[] uvs = new float[0];
		public short[] triangles = new short[0];

		public FloatBuffer packedData;
		public ShortBuffer drawListBuffer;

		public int getVertexPackedDataSize() {
			int dataSize = POSITION_DATA_SIZE;
			if (hasNormals()) {
				dataSize += NORMAL_DATA_SIZE;
			}
			if (hasTangents()) {
				dataSize += TANGENT_DATA_SIZE;
			}
			if (hasUVs()) {
				dataSize += UV_DATA_SIZE;
			}
			// Adds padding if needed to reach a multiple of 4 bytes.
			dataSize += (dataSize % 4);
			return dataSize;
		}

		public boolean hasNormals() {
			return this.normals.length > 0;
		}

		public boolean hasTangents() {
			return this.tangents.length > 0;
		}

		public boolean hasUVs() {
			return this.uvs.length > 0;
		}

		public void build() {
	    	final int vertexCount = vertices.length / POSITION_DATA_SIZE;
	    	final int dataSize = getVertexPackedDataSize();

	    	final float[] vertexData = new float[vertexCount * dataSize];
	    	for (int i = 0; i < vertexCount; i++) {
	    		int writeCursor = i * dataSize;

	    		System.arraycopy(vertices, POSITION_DATA_SIZE * i, vertexData, writeCursor, POSITION_DATA_SIZE);
	    		writeCursor += POSITION_DATA_SIZE;

	    		if (hasNormals()) {
	    			System.arraycopy(normals, NORMAL_DATA_SIZE * i, vertexData, writeCursor, NORMAL_DATA_SIZE);
		    		writeCursor += NORMAL_DATA_SIZE;
	    		}
	    		if (hasTangents()) {
	    			System.arraycopy(tangents, TANGENT_DATA_SIZE * i, vertexData, writeCursor, TANGENT_DATA_SIZE);
		    		writeCursor += TANGENT_DATA_SIZE;
	    		}
	    		if (hasUVs()) {
	    			System.arraycopy(uvs, UV_DATA_SIZE * i, vertexData, writeCursor, UV_DATA_SIZE);
		    		writeCursor += UV_DATA_SIZE;
	    		}
	    	}

	    	final ByteBuffer bb = ByteBuffer.allocateDirect(vertexCount * dataSize * Float.SIZE / 8);
	    	bb.order(ByteOrder.nativeOrder());
	    	this.packedData = bb.asFloatBuffer();
	    	this.packedData.put(vertexData);
	    	this.packedData.position(0);

	    	final ByteBuffer bbDrawList = ByteBuffer.allocateDirect(triangles.length * Short.SIZE / 8);
	    	bbDrawList.order(ByteOrder.nativeOrder());
	    	this.drawListBuffer = bbDrawList.asShortBuffer();
	    	this.drawListBuffer.put(triangles);
	    	this.drawListBuffer.position(0);
		}

		public void recomputeNormals() {
			final int vertexCount = this.vertices.length / POSITION_DATA_SIZE;

			if (this.normals.length / NORMAL_DATA_SIZE != vertexCount) {
				// Not enough place -- or too much -- to store all the coordinates
				// of the normals so we create a new array.
				this.normals = new float[vertexCount * NORMAL_DATA_SIZE];
			}

			final Vector3 p0 = Vector3.getNewZeroVec();
			final Vector3 p1 = Vector3.getNewZeroVec();
			final Vector3 p2 = Vector3.getNewZeroVec();

			final Vector3 v10 = Vector3.getNewZeroVec();
			final Vector3 v12 = Vector3.getNewZeroVec();
			final Vector3 normal = Vector3.getNewZeroVec();

			for (int i = 0, iEnd = this.triangles.length; i < iEnd; i += 3) {
				final int p0Index = this.triangles[i    ];
				final int p1Index = this.triangles[i + 1];
				final int p2Index = this.triangles[i + 2];

				int startVertex = POSITION_DATA_SIZE * p0Index;
				p0.setValues(
						this.vertices[startVertex    ],
						this.vertices[startVertex + 1],
						this.vertices[startVertex + 2]);
				p0.mulSelf(1.0f / this.vertices[startVertex + 3]);
				startVertex = POSITION_DATA_SIZE * p1Index;
				p1.setValues(
						this.vertices[startVertex    ],
						this.vertices[startVertex + 1],
						this.vertices[startVertex + 2]);
				p1.mulSelf(1.0f / this.vertices[startVertex + 3]);
				startVertex = POSITION_DATA_SIZE * p2Index;
				p2.setValues(
						this.vertices[startVertex    ],
						this.vertices[startVertex + 1],
						this.vertices[startVertex + 2]);
				p2.mulSelf(1.0f / this.vertices[startVertex + 3]);

				v10.setValues(p0);
				v10.minusSelf(p1);

				v12.setValues(p2);
				v12.minusSelf(p1);

				Vector3.cross(normal, v12, v10);
				normal.normalizeSelf();

				int startNormal = NORMAL_DATA_SIZE * p0Index;
				this.normals[startNormal    ] = normal.getX();
				this.normals[startNormal + 1] = normal.getY();
				this.normals[startNormal + 2] = normal.getZ();
				startNormal = NORMAL_DATA_SIZE * p1Index;
				this.normals[startNormal    ] = normal.getX();
				this.normals[startNormal + 1] = normal.getY();
				this.normals[startNormal + 2] = normal.getZ();
				startNormal = NORMAL_DATA_SIZE * p2Index;
				this.normals[startNormal    ] = normal.getX();
				this.normals[startNormal + 1] = normal.getY();
				this.normals[startNormal + 2] = normal.getZ();
			}
		}

		public void recomputeTangents() {
			final int vertexCount = this.vertices.length / POSITION_DATA_SIZE;

			if (this.tangents.length / TANGENT_DATA_SIZE != vertexCount) {
				// Not enough place -- or too much -- to store all the coordinates
				// of the tangents so we create a new array.
				this.tangents = new float[vertexCount * TANGENT_DATA_SIZE];
			}

			// Notice: these two creations also set values to 0.
			final float[] tan0 = new float[vertexCount * TANGENT_DATA_SIZE];
			final float[] tan1 = new float[vertexCount * TANGENT_DATA_SIZE];

			final Vector3 p0 = Vector3.getNewZeroVec();
			final Vector3 p1 = Vector3.getNewZeroVec();
			final Vector3 p2 = Vector3.getNewZeroVec();

			final Vector3 tmp0 = Vector3.getNewZeroVec();
			final Vector3 tmp1 = Vector3.getNewZeroVec();

			final Vector3 tmpP0 = Vector3.getNewZeroVec();
			final Vector3 tmpP1 = Vector3.getNewZeroVec();

			final Vector3 sDir = Vector3.getNewZeroVec();
			final Vector3 tDir = Vector3.getNewZeroVec();

			for (int i = 0, iEnd = this.triangles.length; i < iEnd; i += 3) {
				final int p0Index = this.triangles[i    ];
				final int p1Index = this.triangles[i + 1];
				final int p2Index = this.triangles[i + 2];

				int startVertex = POSITION_DATA_SIZE * p0Index;
				p0.setValues(
						this.vertices[startVertex    ],
						this.vertices[startVertex + 1],
						this.vertices[startVertex + 2]);
				p0.mulSelf(1.0f / this.vertices[startVertex + 3]);
				startVertex = POSITION_DATA_SIZE * p1Index;
				p1.setValues(
						this.vertices[startVertex    ],
						this.vertices[startVertex + 1],
						this.vertices[startVertex + 2]);
				p1.mulSelf(1.0f / this.vertices[startVertex + 3]);
				startVertex = POSITION_DATA_SIZE * p2Index;
				p2.setValues(
						this.vertices[startVertex    ],
						this.vertices[startVertex + 1],
						this.vertices[startVertex + 2]);
				p2.mulSelf(1.0f / this.vertices[startVertex + 3]);

				Vector3.minus(tmpP0, p1, p0);
				Vector3.minus(tmpP1, p2, p0);

				final float w0U = this.uvs[UV_DATA_SIZE * p0Index    ];
				final float w0V = this.uvs[UV_DATA_SIZE * p0Index + 1];
				final float w1U = this.uvs[UV_DATA_SIZE * p1Index    ];
				final float w1V = this.uvs[UV_DATA_SIZE * p1Index + 1];
				final float w2U = this.uvs[UV_DATA_SIZE * p2Index    ];
				final float w2V = this.uvs[UV_DATA_SIZE * p2Index + 1];

				final float s0 = w1U - w0U;
				final float s1 = w2U - w0U;
				final float t0 = w1V - w0V;
				final float t1 = w2V - w0V;

			    final float divInR = s0 * t1 - s1 * t0;
			    final float r = (divInR == 0.0f ? 0.0f : 1.0f / divInR);
				// sDir = (t1 * tmp0 - t0 * tmp1) * r
				tmp0.setValues(tmpP0);
				tmp0.mulSelf(t1);
				tmp1.setValues(tmpP1);
				tmp1.mulSelf(t0);
				Vector3.minus(sDir, tmp0, tmp1);
				sDir.mulSelf(r);
				// tDir = (s0 * tmp1 - s1 * tmp0) * r
				tmp0.setValues(tmpP1);
				tmp0.mulSelf(s0);
				tmp1.setValues(tmpP0);
				tmp1.mulSelf(s1);
				Vector3.minus(tDir, tmp0, tmp1);
				tDir.mulSelf(r);

				for (int j = 0; j < 3; j++) {
					tan0[p0Index * TANGENT_DATA_SIZE + j] += sDir.getComponent(j);
					tan0[p1Index * TANGENT_DATA_SIZE + j] += sDir.getComponent(j);
					tan0[p2Index * TANGENT_DATA_SIZE + j] += sDir.getComponent(j);
				}
				for (int j = 0; j < 3; j++) {
					tan1[p0Index * TANGENT_DATA_SIZE + j] += tDir.getComponent(j);
					tan1[p1Index * TANGENT_DATA_SIZE + j] += tDir.getComponent(j);
					tan1[p2Index * TANGENT_DATA_SIZE + j] += tDir.getComponent(j);
				}
			}

			final Vector3 tmpNormal = Vector3.getNewZeroVec();
			final Vector3 tmpTangent = Vector3.getNewZeroVec();
			final Vector3 finalTangent = Vector3.getNewZeroVec();

		    for (int i = 0; i < vertexCount; i++) {
		    	final int startTmpNormal = i * NORMAL_DATA_SIZE;
		    	tmpNormal.setValues(
		    			this.normals[startTmpNormal    ],
		    			this.normals[startTmpNormal + 1],
		    			this.normals[startTmpNormal + 2]);
		    	final int startTmpTangent = i * TANGENT_DATA_SIZE;
		    	tmpTangent.setValues(
		    			tan0[startTmpTangent    ],
		    			tan0[startTmpTangent + 1],
		    			tan0[startTmpTangent + 2]);

		    	// Gram-Schmidt orthogonalize
		        // finalTangent = (t - n * Dot(n, t)).Normalize();
		    	finalTangent.setValues(tmpNormal);
		    	final float dotN_T = Vector3.dot(tmpNormal, tmpTangent);
		    	finalTangent.mulSelf(-dotN_T);
		    	finalTangent.addSelf(tmpTangent);
		    	finalTangent.normalizeSelf();

		    	this.tangents[startTmpTangent    ] = finalTangent.getX();
		    	this.tangents[startTmpTangent + 1] = finalTangent.getY();
		    	this.tangents[startTmpTangent + 2] = finalTangent.getZ();

		        // Calculate handedness
		        // tangent.w = (Dot(Cross(n, t), tan1[i]) < 0.0F) ? -1.0f : 1.0f;
		    	//           = (Dot(tmp0, tmp1)  < 0.0f) ? -1.0f : 1.0f;
		    	Vector3.cross(tmp0, tmpNormal, tmpTangent);
		    	tmp1.setValues(
		    			tan1[startTmpTangent    ],
		    			tan1[startTmpTangent + 1],
		    			tan1[startTmpTangent + 2]);
		    	this.tangents[startTmpTangent + 3] = (
		    			Vector3.dot(tmp0, tmp1) < 0.0f ? -1.0f : 1.0f);
		    }
		}
	}

	private final ArrayList<SubMeshData> subMeshDataSet = new ArrayList<SubMeshData>();

	protected final SubMeshData getOrCreateSubMeshData(int subMeshId) {
		for (int i = this.subMeshDataSet.size(), iEnd = subMeshId + 1; i < iEnd; i++) {
			this.subMeshDataSet.add(new SubMeshData());
		}
		return this.subMeshDataSet.get(subMeshId);
	}

	protected static float[] copyToOrCreateArray(float[] srcArray, float[] dstArray) {
		if (dstArray.length != srcArray.length) {
			dstArray = srcArray.clone();
		} else {
			System.arraycopy(srcArray, 0, dstArray, 0, srcArray.length);
		}
		return dstArray;
	}

	protected static short[] copyToOrCreateArray(short[] srcArray, short[] dstArray) {
		if (dstArray.length != srcArray.length) {
			dstArray = srcArray.clone();
		} else {
			System.arraycopy(srcArray, 0, dstArray, 0, srcArray.length);
		}
		return dstArray;
	}

	/**
	 * Gets the lowest id of a sub-mesh with the provided name.
	 * @param name
	 * @return the lowest possible id (it might be <code>0</code>), or
	 *   <code>-1</code> if none was found.
	 */
	public final int getSubMeshIdByName(String name) {
		int ret = -1;
		for (int i = 0, iEnd = this.subMeshDataSet.size(); i < iEnd && ret == -1; i++) {
			if (name == this.subMeshDataSet.get(i).name) {
				ret = i;
			}
		}
		return ret;
	}

	public final void setSubMeshName(String name, int subMeshId) {
		final SubMeshData meshData = getOrCreateSubMeshData(subMeshId);
		meshData.name = name;
	}

	public final String getSubMeshName(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).name;
	}

	public final void setVertices(float[] vertices, int subMeshId) {
		final SubMeshData meshData = getOrCreateSubMeshData(subMeshId);
		meshData.vertices = copyToOrCreateArray(vertices, meshData.vertices);
	}

	public final float[] getSharedVertices(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).vertices;
	}

	public final float[] getVertices(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).vertices.clone();
	}

	public final boolean hasNormals(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).hasNormals();
	}

	public final void setNormals(float[] normals, int subMeshId) {
		final SubMeshData meshData = getOrCreateSubMeshData(subMeshId);
		meshData.normals = copyToOrCreateArray(normals, meshData.normals);
	}

	public final float[] getSharedNormals(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).normals;
	}

	public final float[] getNormals(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).normals.clone();
	}

	public final boolean hasTangents(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).hasTangents();
	}

	public final void setTangents(float[] tangents, int subMeshId) {
		final SubMeshData meshData = getOrCreateSubMeshData(subMeshId);
		meshData.tangents = copyToOrCreateArray(tangents, meshData.tangents);
	}

	public final float[] getSharedTangents(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).tangents;
	}

	public final float[] getTangents(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).tangents.clone();
	}

	public final boolean hasUVs(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).hasUVs();
	}

	public final void setUVs(float[] uvs, int subMeshId) {
		final SubMeshData meshData = getOrCreateSubMeshData(subMeshId);
		meshData.uvs = copyToOrCreateArray(uvs, meshData.uvs);
	}

	public final float[] getSharedUVs(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).uvs;
	}

	public final float[] getUVs(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).uvs.clone();
	}

	public final void setTriangles(short[] triangles, int subMeshId) {
		final SubMeshData meshData = getOrCreateSubMeshData(subMeshId);
		meshData.triangles = copyToOrCreateArray(triangles, meshData.triangles);
	}

	public final short[] getSharedTriangles(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).triangles;
	}

	public final short[] getTriangles(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).triangles;
	}

	/**
	 * Recomputes the normals. This method can be used only once the method
	 * <code>setTriangles</code> has been called because it will use triangles.
	 * In addition, this method will take care of creating a new array to store
	 * the coordinates of the normals if the old one does not have the exact size.
	 * @param subMeshId
	 */
	public final void recomputeNormals(int subMeshId) {
		this.subMeshDataSet.get(subMeshId).recomputeNormals();
	}

	/**
	 * Recomputes the tangents. This method can be used only once the methods
	 * <code>setNormals</code>, <code>setUVs</code> and <code>setTriangles</code>
	 * have been called because normals, UV coordinates and triangles are used.
	 * In addition, this method will take care of creating a new array to store
	 * the coordinates of the tangents if the old one does not have the exact size.
	 * @param subMeshId
	 */
	public final void recomputeTangents(int subMeshId) {
		this.subMeshDataSet.get(subMeshId).recomputeTangents();
	}

	public void clear() {
		this.subMeshDataSet.clear();
	}

	public void recomputeAllNormals() {
		for (int i = 0, iEnd = this.subMeshDataSet.size(); i < iEnd; i++) {
			this.subMeshDataSet.get(i).recomputeNormals();
		}
	}

	public void recomputeAllTangents() {
		for (int i = 0, iEnd = this.subMeshDataSet.size(); i < iEnd; i++) {
			this.subMeshDataSet.get(i).recomputeTangents();
		}
	}

	public void build() {
		for (int i = 0, iEnd = this.subMeshDataSet.size(); i < iEnd; i++) {
			this.subMeshDataSet.get(i).build();
		}
	}

	public final int getNbOfSubMeshes() {
		return this.subMeshDataSet.size();
	}

	public int getVertexPackedDataSize(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).getVertexPackedDataSize();
	}

	public final int getVertexCount(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).vertices.length / POSITION_DATA_SIZE;
	}

	public final int getVertexCoordinatesCount(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).vertices.length;
	}

	public final FloatBuffer getSharedVertexPackedData(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).packedData;
	}

	public final int getDrawListCount(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).triangles.length;
	}

	public final ShortBuffer getSharedDrawListBuffer(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).drawListBuffer;
	}
}
