package com.laguerre.kan;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;

import com.laguerre.kan.MeshImporter;
import com.laguerre.kan.MeshImporter.Settings;

public class ObjImporter implements MeshImporter.IMeshImporterImpl {
	public static class UnexpectedNumberOfValuesException extends RuntimeException {
		private static final long serialVersionUID = 6867346167225379981L;

		public UnexpectedNumberOfValuesException(String msg) {
			super(msg);
		}
	}

	public static class TooHightIndexException extends RuntimeException {
		private static final long serialVersionUID = -7506392685572998497L;

		public TooHightIndexException(String msg) {
			super(msg);
		}
	}

	private Settings settings;

	@Override
	public void setSettings(Settings settings) {
		this.settings = settings;
	}

	@Override
	public void importToTarget(InputStream inputStream, SceneObject sobTarget) {
		Mesh mesh = sobTarget.getAnyComponent(Mesh.class);
		if (mesh == null) {
			 mesh = sobTarget.addComponent(Mesh.class);
		}
		loadContent(inputStream, mesh);
	}

	protected void loadContent(InputStream inputStream, Mesh mesh) {
		final InputStreamReader inputStreamReader = new InputStreamReader(
				inputStream);
		final BufferedReader bufferedReader = new BufferedReader(inputStreamReader, 8192);
		try {
			final Renderer renderer = mesh.getSharedRenderer();

			// Removes any previous data that represents this mesh.
			mesh.clear();

			// Will contain read data.
			final ArrayList<Float> readVertexData = new ArrayList<Float>();
			final ArrayList<Float> readNormalData = new ArrayList<Float>();
			final ArrayList<Float> readUVData = new ArrayList<Float>();

			// Will contain data built dynamically
			String subMeshName = "";
			int subMeshId = 0;
			// Below, each index contains values for the same vertex (coordinates,
			// normals and uv).
			final ArrayList<Float> finalVertexData = new ArrayList<Float>();
			final ArrayList<Float> finalNormalData = new ArrayList<Float>();
			final ArrayList<Float> finalUVData = new ArrayList<Float>();
			final ArrayList<Integer> finalFaceData = new ArrayList<Integer>();

			String currentLine = null;
			while ((currentLine = bufferedReader.readLine()) != null) {
				currentLine = currentLine.trim();
				final String[] lineContent = currentLine.split("\\s+");

				if (lineContent.length == 0 || lineContent[0].startsWith("#")) {
					// Nothing to do...
					continue;
				} else if (lineContent[0].equals("v")) {
					if (lineContent.length < 4) {
						throw new UnexpectedNumberOfValuesException("A vertex did not provide at least 3 values");
					}
					for (int i = 1; i < 4; i++) {
						readVertexData.add(this.settings.globalScale * Float.parseFloat(lineContent[i]));
					}
					if (lineContent.length >= 5) {
						// "w" is provided.
						readVertexData.add(Float.parseFloat(lineContent[4]));
					} else {
						// Sets "w" as 1.0
						readVertexData.add(1.0f);
					}
					// Assumes that vertices was the first data to be loaded so
					// memory is reserved for further loading of normals and UV.
					// It will be a waste if no normals are provided but much
					// faster in the opposite case...
					final int nbVertices = readVertexData.size() / 4;
					readNormalData.ensureCapacity(nbVertices * 3);
					readUVData.ensureCapacity(nbVertices * 2);
				} else if (lineContent[0].equals("vn")) {
					if (lineContent.length != 4) {
						throw new UnexpectedNumberOfValuesException("A normal did not provide 3 values");
					}
					for (int i = 1; i < 4; i++) {
						readNormalData.add(Float.parseFloat(lineContent[i]));
					}
				} else if (lineContent[0].equals("vt")) {
					if (lineContent.length != 3 && lineContent.length != 4) {
						throw new UnexpectedNumberOfValuesException("A UV did not provide 2 values");
					}
					if (lineContent.length == 4) {
						final float tCoord = Float.parseFloat(lineContent[3]);
						if (!Utils.fuzzyEquals(0.0f, tCoord)) {
							throw new UnexpectedNumberOfValuesException("UV only supports 2 values, or the third must equal to 0");
						}
					}
					for (int i = 1; i < 3; i++) {
						readUVData.add(Float.parseFloat(lineContent[i]));
					}
					if (this.settings.swappedUV) {
						final int readUVDataSize = readUVData.size();
						final float oldUComponent = readUVData.get(readUVDataSize - 2);
						readUVData.set(readUVDataSize - 2, readUVData.get(readUVDataSize - 1));
						readUVData.set(readUVDataSize - 1, oldUComponent);
					}
				} else if (lineContent[0].equals("f")) {
					if (lineContent.length < 4) {
						throw new UnexpectedNumberOfValuesException("A face did not provide at least 3 vertices");
					}

					for (int i = 2; i + 1 < lineContent.length; i++) {
						// Adds p (1), it is the beginning of a triangle fan.
						addVertexData(
								lineContent[1],
								readVertexData,
								readNormalData,
								readUVData,
								finalVertexData,
								finalNormalData,
								finalUVData);
						finalFaceData.add(finalFaceData.size());

						// Adds p (i)
						addVertexData(
								lineContent[i],
								readVertexData,
								readNormalData,
								readUVData,
								finalVertexData,
								finalNormalData,
								finalUVData);
						finalFaceData.add(finalFaceData.size());

						// Adds p (i + 1)
						addVertexData(
								lineContent[i + 1],
								readVertexData,
								readNormalData,
								readUVData,
								finalVertexData,
								finalNormalData,
								finalUVData);
						finalFaceData.add(finalFaceData.size());
					}
				} else if (lineContent[0].equals("g") ||
						lineContent[0].equals("usemtl")) {
					// Any mesh data loaded previously? If true, builds the mesh.
					if (!finalVertexData.isEmpty() && !finalFaceData.isEmpty()) {
						setObjMeshData(mesh, subMeshId, subMeshName,
								finalVertexData, finalNormalData, finalUVData, finalFaceData);
						subMeshId++;
						// Ensures that we aren't keeping garbage.
						finalVertexData.clear();
						finalNormalData.clear();
						finalUVData.clear();
						finalFaceData.clear();
					}

					if (lineContent[0].equals("g")) {
						// The name of the new mesh we're beginning to load.
						if (lineContent.length > 1) {
							// Number 2 for "g ".
							subMeshName = currentLine.substring(2).trim();
						}
					} else {
						// Material name is declared.
						if (renderer != null && lineContent.length > 1) {
							Material matSubMesh = renderer.getSharedMaterial(subMeshId);
							if (matSubMesh == null) {
								matSubMesh = new Material();
								renderer.setMaterial(matSubMesh, subMeshId);
							}
							// Number 7 for "usemtl ".
							final String materialName = currentLine.substring(7).trim();
							matSubMesh.setMaterialName(materialName);
						}
					}
				}
			}

			if (!finalVertexData.isEmpty() && !finalFaceData.isEmpty()) {
				setObjMeshData(mesh, subMeshId, subMeshName,
						finalVertexData, finalNormalData, finalUVData, finalFaceData);
			}

			// And builds the mesh...
			mesh.build();
		} catch (IOException e) {
		}
	}

	protected void addVertexData(
			final String vertexDataStr,
			final ArrayList<Float> readVertexData,
			final ArrayList<Float> readNormalData,
			final ArrayList<Float> readUVData,
			final ArrayList<Float> finalVertexData,
			final ArrayList<Float> finalNormalData,
			final ArrayList<Float> finalUVData) {
		final String[] faceContent = vertexDataStr.split("/");
		boolean onlyVertexIndex = false;
		boolean normalProvided = false;
		boolean uvProvided = false;
		int vertexIndex;
		int normalIndex = -1 ;
		int uvIndex = -1;
		if (faceContent.length == 1) {
			onlyVertexIndex = true;
			// Only vertex indexes are given.
			vertexIndex = Integer.parseInt(vertexDataStr);
		} else {
			if (faceContent.length == 2) {
				// Format: v/vt.
				vertexIndex = Integer.parseInt(faceContent[0]);
				// Always a normal.
				uvIndex = Integer.parseInt(faceContent[1]);
				uvProvided = true;
			} else {
				// Format: v/vt/vn.
				// An uv can be ignored with a slash, e.g v//vn.
				vertexIndex = Integer.parseInt(faceContent[0]);
				// Always a normal.
				normalIndex = Integer.parseInt(faceContent[2]);
				normalProvided = true;
				// UV also?
				if (faceContent[1].length() > 0) {
					uvIndex = Integer.parseInt(faceContent[1]);
					uvProvided = true;
				}
			}
		}
		if (vertexIndex < 0) {
			// negative means that it is a reference to the just-added vertices.
			vertexIndex = readVertexData.size() + vertexIndex;
		} else {
			// First index is 0 for us, not 1.
			vertexIndex--;
		}

		// Copies vertex coordinates.
		for (int j = Mesh.POSITION_DATA_SIZE * vertexIndex, jEnd = j + Mesh.POSITION_DATA_SIZE; j < jEnd; j++) {
			finalVertexData.add(readVertexData.get(j));
		}

		if (onlyVertexIndex) {
			// No index provided? But maybe some have been loaded
			// and thus index is also the vertex index...
			if (vertexIndex + Mesh.NORMAL_DATA_SIZE < readNormalData.size()) {
				for (int j = Mesh.NORMAL_DATA_SIZE * vertexIndex, jEnd = j + Mesh.NORMAL_DATA_SIZE; j < jEnd; j++) {
					finalNormalData.add(readNormalData.get(j));
				}
			}
			// And for UV?
			if (vertexIndex + Mesh.UV_DATA_SIZE < readUVData.size()) {
				for (int j = Mesh.UV_DATA_SIZE * vertexIndex, jEnd = j + Mesh.UV_DATA_SIZE; j < jEnd; j++) {
					finalUVData.add(readUVData.get(j));
				}
			}
		} else {
			// Copies normal coordinates, if any.
			if (normalProvided) {
				// First index is 0 for us, not 1.
				normalIndex--;
				for (int j = Mesh.NORMAL_DATA_SIZE * normalIndex, jEnd = j + Mesh.NORMAL_DATA_SIZE; j < jEnd; j++) {
					finalNormalData.add(readNormalData.get(j));
				}
			}
			// Copies uv coordinates, if any.
			if (uvProvided) {
				// First index is 0 for us, not 1.
				uvIndex--;
				for (int j = Mesh.UV_DATA_SIZE * uvIndex, jEnd = j + Mesh.UV_DATA_SIZE; j < jEnd; j++) {
					finalUVData.add(readUVData.get(j));
				}
			}
		}
	}

	protected void setObjMeshData(
			final Mesh mesh,
			final int subMeshId,
			final String subMeshName,
			final ArrayList<Float> vertexData,
			final ArrayList<Float> normalData,
			final ArrayList<Float> uvData,
			final ArrayList<Integer> faceData) {
		float[] floatData = new float[vertexData.size()];
		for (int i = 0, iEnd = vertexData.size(); i < iEnd; i++) {
			floatData[i] = vertexData.get(i);
		}
		mesh.setVertices(floatData, subMeshId);

		short[] shortData = new short[faceData.size()];
		for (int i = 0, iEnd = faceData.size(); i < iEnd; i++) {
			//
			final int vertexIndex = faceData.get(i);
			if (vertexIndex > Short.MAX_VALUE) {
				throw new TooHightIndexException("Got index \"" + vertexIndex +
						"\", which is too big (> " + Short.MAX_VALUE + ")");
			}
			shortData[i] = (short) vertexIndex;
		}
		mesh.setTriangles(shortData, subMeshId);

		switch (this.settings.normalImportMode) {
		case IMPORT_FROM_FILE:
			if (!normalData.isEmpty()) {
				floatData = new float[normalData.size()];
				for (int i = 0, iEnd = normalData.size(); i < iEnd; i++) {
					floatData[i] = normalData.get(i);
				}
				mesh.setNormals(floatData, subMeshId);
			}
			break;
		case CALCULATE:
			mesh.recomputeNormals(subMeshId);
			break;
		case NONE:
		default:
			break;
		}

		if (!uvData.isEmpty()) {
			floatData = new float[uvData.size()];
			for (int i = 0, iEnd = uvData.size(); i < iEnd; i++) {
				floatData[i] = uvData.get(i);
			}
			mesh.setUVs(floatData, subMeshId);
		}

		switch (this.settings.tangentImportMode) {
		case IMPORT_FROM_FILE:
			// Not supported by this file format.
			break;
		case CALCULATE:
			mesh.recomputeTangents(subMeshId);
			break;
		case NONE:
		default:
			break;
		}
	}
}
